#include "planner.h"
#include "node.h"
#include <iostream>
#include <vector>
using namespace std;
using namespace __gnu_cxx;

Planner::Planner(std::vector<Action> actions, Literals init, Literals goal){
	_init = init;
	_goal = goal;
	_start = new StartNode(_init);
	_finish = new EndNode(_goal);
	addNode(_start);

	for(vector<Action>::iterator action = actions.begin(); action != actions.end(); ++action){
		const Literals& effects = action->effects(0);
		for (Literals::const_iterator effect = effects.begin(); effect != effects.end(); ++effect){
			cerr << "Adding effect: '" << *effect << "', action: " << action->name() << endl;
			_actions[*effect] = *action;
		}
	}
	cerr << "Number of actions: " << _actions.size() << endl;
	makePlan(_finish);
}

Planner::~Planner(){
	cerr << "Deleting " << _addedNodes.size() << " nodes" << endl;
	for (vector<Node*>::iterator node = _addedNodes.begin(); node != _addedNodes.end(); ++node){
		cerr << "Deleting node " << (*node)->action().name() << endl;
		delete *node;
	}
}


void Planner::makePlan(Node* node){
	addNode(node);

	cerr << "Fetching preconditions for action: " << node->action().name() << ".. ";
	const Preconditions& preconds = node->action().preconditions();
	cerr << "done" << endl;


	if (preconds.size() == 0){
		cerr << "Found no preconds" << endl;
		_start->addChild(node);
	}else{
		for (Preconditions::const_iterator precond = preconds.begin(); precond != preconds.end(); ++precond){
			cerr << "Looking for: '" << precond->first << "'" <<  endl;
			hash_map<string,Node*>::iterator addedNode = _addedEffects.find(precond->first);
			if(addedNode != _addedEffects.end()){
				cerr << "Using already added node" << endl;
				addedNode->second->addChild(node);
			}else {
				hash_map<string, Action>::iterator action = _actions.find(precond->first);
				if (action != _actions.end()){
					cerr << "Adding new node" << endl;
					Node* newnode = new Node(action->second);
					newnode->addChild(node);
					makePlan(newnode);
				}else if (precond->second){
					cerr << "Action with effect: " << precond->first << " not found!" << endl;
					cerr << "This is a hard precondition, so this action and the children can't be executed." << endl;
					return;
				}else{
					cerr << "Action with effect: " << precond->first << " not found!" << endl;
					cerr << "This is a soft precondition, so we will continue" << endl;
					node->satisfyCondition(precond->first);
				}
			}
		}
	}
}

void Planner::addNode(Node* node){
	cerr << "Adding node for action: " << node->action().name() << endl;
	const Literals& effects = node->action().effects(0);
	cerr << "Number of effects: " << effects.size() << endl;
	_addedNodes.push_back(node);

	for (Literals::const_iterator effect = effects.begin(); effect != effects.end(); ++effect){
		cerr << "Adding node for effect: " << *effect << endl;
		_addedEffects[*effect] = node;
	}
}


void Planner::execute(){
	_start->execute(Literals());
	cerr << "Number of nodes: " << _addedNodes.size() << endl;
	for (vector<Node*>::iterator node = _addedNodes.begin(); node != _addedNodes.end(); ++node){
		if ((*node)->executed()){
			//remove stuff
		}
		delete *node;
	}
	_addedNodes.clear();
	cerr << "Number of nodes left: " << _addedNodes.size() << endl;
}
